import os
import sys
from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from pydantic import ValidationError
from sqlalchemy import create_engine, event
from sqlalchemy.orm import sessionmaker

from argos.logging import logger
from argos.server import models, routes, queries
from argos.server.settings import read_yaml_config


def get_application() -> FastAPI:
    """Spawn Argos FastAPI server"""
    appli = FastAPI()
    config_file = os.environ["ARGOS_YAML_FILE"]

    config = read_config(config_file)

    # Config is the argos config object (built from yaml)
    appli.state.config = config

    appli.add_event_handler(
        "startup",
        create_start_app_handler(appli),
    )
    appli.add_event_handler(
        "shutdown",
        create_stop_app_handler(appli),
    )
    appli.include_router(routes.api, prefix="/api")
    appli.include_router(routes.views)

    static_dir = Path(__file__).resolve().parent / "static"

    appli.mount("/static", StaticFiles(directory=static_dir), name="static")
    return appli


def create_start_app_handler(appli):
    """Warmup the server:
    setup database connection
    """

    async def _get_db():
        setup_database(appli)

        db = await connect_to_db(appli)

        tasks_count = await queries.count_tasks(db)
        if tasks_count == 0:
            logger.warning(
                "There is no tasks in the database. "
                'Please launch the command "argos server reload-config"'
            )

        return db

    return _get_db


async def connect_to_db(appli):
    appli.state.db = appli.state.SessionLocal()
    return appli.state.db


def create_stop_app_handler(appli):
    """Gracefully shutdown the server:
    close database connection.
    """

    async def stop_app():
        appli.state.db.close()

    return stop_app


def read_config(yaml_file):
    try:
        config = read_yaml_config(yaml_file)
        return config
    except ValidationError as err:
        logger.error("Errors where found while reading configuration:")
        for error in err.errors():
            logger.error("%s is %s", error["loc"], error["type"])
        sys.exit(1)


def setup_database(appli):
    config = appli.state.config
    db_url = str(config.general.db.url)
    logger.debug("Using database URL %s", db_url)
    # For sqlite, we need to add connect_args={"check_same_thread": False}
    if config.general.env == "production" and db_url.startswith("sqlite:////tmp"):
        logger.warning("Using sqlite in /tmp is not recommended for production")

    extra_settings = {}
    if config.general.db.pool_size:
        extra_settings.setdefault("pool_size", config.general.db.pool_size)

    if config.general.db.max_overflow:
        extra_settings.setdefault("max_overflow", config.general.db.max_overflow)

    engine = create_engine(db_url, **extra_settings)

    def _fk_pragma_on_connect(dbapi_con, con_record):
        dbapi_con.execute("pragma foreign_keys=ON")

    if db_url.startswith("sqlite:///"):
        event.listen(engine, "connect", _fk_pragma_on_connect)

    appli.state.SessionLocal = sessionmaker(
        autocommit=False, autoflush=False, bind=engine
    )
    appli.state.engine = engine
    models.Base.metadata.create_all(bind=engine)


app = get_application()
